/* macbhex.c -- simple binhex decoding routine */
/* (C) Copyright 1993,1994 by Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Carnegie
 * Mellon University not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  Carnegie Mellon University makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/* modified for maish 1999 Shinichi Honma (Rick) */
 
#include <stdio.h>
#include <stdlib.h>
#include "maish.h"

extern char	file_outname[STRBUF],			/* o̓t@C */
			base[94],						/* WJfBNg */
			exfile[STRBUF];					/* owt@C */

char binhex_decode[256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, -1, -1,
	13, 14, 15, 16, 17, 18, 19, -1, 20, 21, -2, -1, -1, -1, -1, -1,
	22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, -1,
	37, 38, 39, 40, 41, 42, 43, -1, 44, 45, 46, 47, -1, -1, -1, -1,
	48, 49, 50, 51, 52, 53, 54, -1, 55, 56, 57, 58, 59, 60, -1, -1,
	61, 62, 63, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
};
#define BHEXVAL(c) (binhex_decode[(unsigned char) c])

#define	BBUFSIZE	32768

#define STATE_START  0
#define STATE_FNAME  1
#define STATE_HEADER 2
#define STATE_HCRC   3
#define STATE_DFORK  4
#define STATE_DONE   8
#define STATE_ERROR  9

typedef struct {
	unsigned char	filler[10];
	long	dlen, rlen;
} binhex_header;

typedef struct binhex_state {
	long 	ct;					/* bbuf_counter */
	unsigned char	*buf;		/* bbuf */
	long 	bmax;				/* bbuf_max */
	short	state;				/* current state */
	long 	count;				/* generic counter */
	FILE	*outfile;			/* output file */
	long	namelen;			/* namelen */
	binhex_header	head;		/* buffer for header */
} binhex_state;

/* process a binhex character
 */
static void binhex_process(binhex_state *bp)
{
	long	len= bp->ct;
	unsigned char *ptr= bp->buf;

	do {
		unsigned char c = *ptr++;

		/* handle state */
		switch (bp->state) {
			case STATE_START:
				bp->state = STATE_FNAME;
				bp->count = 0;						/* JE^ */
				bp->namelen = (c & 63);				/* t@C(63ȉ) */
				break;
			case STATE_FNAME:
				file_outname[bp->count] = c;		/* t@CPbyteo */
				if (bp->count++ >= bp->namelen) {	/* t@CɒB */
					file_outname[bp->count] = 0x00;
					if (!*exfile || !p_strcmpi (file_outname, exfile)) {
						bp->state = STATE_HEADER;
						bp->count = 0;				/* JE^ */
					}else {
						bp->state = STATE_DONE;
						return;
					}
				}
				break;
			case STATE_HEADER:						/* wb_1byteo */
				((char *)&(bp->head))[bp->count] = c;
				if (++bp->count == 18) {			/* 18byteɒB玟 */
					bp->state = STATE_HCRC;
					bp->count = 0;
				}
				break;
			case STATE_HCRC:
				if (bp->count++) {					/* CRCǂݎ̂ */
													/* JE^=f[^ */
					if ( bp->count = bp->head.dlen ) {
						char	fullname[STRBUF], *ptr;
						ptr = p_strcpy (fullname, base);
						p_strcpy (ptr, file_outname);
						fputs (file_outname, stderr);
						bp->outfile = fopen (fullname, "wb");
						if (!(bp->outfile)) {			/* I[vs */
							bp->state = STATE_ERROR;
							fputs (" open error\n", stderr);
							return;
						}
						fputc( ' ', stderr );
						bp->state = STATE_DFORK;		/* f[^fR[h */
						bp->bmax= (bp->count > BBUFSIZE)? BBUFSIZE:bp->count;
					} else {
/*						fputs(" is empty", stderr);*/
						bp->state = STATE_DONE;		/* I */
						return;
					}
				}
				break;
			case STATE_DFORK:						/* obt@o */
				if (len >= bp->count) {
					len = bp->count;
					bp->state = STATE_DONE;
				}
				bp->count -= len;
				if( bp->outfile ){
					fputc( 'o', stderr );
					fwrite ( ptr - 1, sizeof (unsigned char), len, bp->outfile);
				}
				bp->bmax= (bp->count > BBUFSIZE)? BBUFSIZE:bp->count;
				return;
			default:
				break;
		}
	} while (--len > 0);
}

static void	bputc (unsigned char chr, binhex_state *bp)
{
	bp->buf[bp->ct++] = chr;
	if (bp->ct == bp->bmax) {
		binhex_process (bp);
		bp->ct = 0;
	}
}

static void	bputcl (unsigned char chr, long len, binhex_state *bp)
{
	while (len) {
		long	max;
		unsigned char	*ptr;

		max = (len + bp->ct > bp->bmax)? (bp->bmax - bp->ct):len;
		ptr = bp->buf + bp->ct;
		len -= max;
		bp->ct+= max;

		while (--max >= 0)	*ptr++ = chr;

		if ( bp->ct == bp->bmax) {
			binhex_process (bp);
			bp->ct = 0;
		}
	}
}


/*
 * decode a binhex file
 *  returns TRUE on fatal error, FALSE for done
 */
int ai_binhex (FILE *rfp, unsigned char *ibuf)
{
	/* global state */
	binhex_state bh;

	unsigned long	octetin, octetbuf;
	unsigned char	*ptr,
					old,
					compress = FALSE,
					obuf[1024];

	/* skip blank lines */
	while (1) {
		if ( !fgets (ibuf, IOBUF, rfp) ) {
			return FALSE;
		}
		if (*ibuf != '\n')	break;						/* ŝ݂̍sƂ΂ */
	}

	/* reset state */
	bh.state = STATE_START;
	bh.outfile = 0x00;
	bh.ct = 0;
	bh.bmax= 1;
	octetin = 4;
	octetbuf = 0;

	ptr = ibuf;
	if (*ptr++ != ':') {								/* 擪':' */
/*		bh.state = STATE_ERROR;
		fputs (" bad start code", stderr);*/
		return TRUE;
	}

	bh.buf = malloc (BBUFSIZE);
	if ( !(bh.buf) ) {
		fputs ("maish : memory exhausted\n", stderr);
		return FALSE;
	}

	/* line reading loop */
	do {
		long	byte = 0;
		unsigned char	*wptr = obuf;
		char	val;

		/* loop through line of binhex charaters */
		do {										/* convert_line */
			/* fill in octetbuf */
			for (; octetin > 0; octetin--) {		/* 4byteP */
				val = BHEXVAL (*ptr++);				/* R[hϊ */
				if (val < 0) {						/* gps? */
					if (val == -2 && (octetin & 3)) {
													/* Ō':'Ȃlߕ */
						octetbuf = octetbuf << (octetin * 6);
						byte = -octetin;
						octetin = 0;
					}
					break;
				}
				octetbuf = octetbuf << 6;			/* 6bitpbN */
				octetbuf |= val;
			}

			if (!octetin) {							/* 4byte܂ */
				*wptr++ = (octetbuf >> 16) & 0xff;
				*wptr++ = (octetbuf >> 8) & 0xff;
				*wptr++ = octetbuf & 0xff;
				octetin = 4;
				octetbuf = 0;
			}

		} while (val >= 0);

		byte += wptr - obuf;
		if (!byte)	break;

		/* handle decoded characters -- run length encoding (rle) detection */
		ptr = obuf;
													/* Lbyte[v */
		while (bh.state < STATE_DONE && --byte >= 0) {
			unsigned char	c = *ptr++;

			/* get character & handle rle */
			if (compress) {							/* k[h */
				if (!c) {							/* 0Ȃ */
					bputc (0x90, &bh);
				} else {
					bputcl (old, c - 1, &bh);		/* O̕c-1o */
				}
				compress = FALSE;					/* k[hI */
			} else {
				if (c == 0x90) {
					compress = TRUE;				/* k[hJn */
				} else {
					old = c;
					bputc (c, &bh);				/* o */
				}
			}
		}
													/* EOFȂ璆~ */
	} while (bh.state <= STATE_DONE && fgets (ptr= ibuf, IOBUF, rfp));

	if (bh.state < STATE_DONE) {					/* łG[łȂ */
		if (bh.ct) {								/* bflush */
			binhex_process (&bh);
		}
	}

	if (bh.outfile) {
		fclose (bh.outfile);
		fputc( '\n', stderr );
	}

	free (bh.buf);

	return (bh.state == STATE_ERROR);
}
